module Data.Louvain where

import Data.List (maximumBy)
import Data.Graph.Inductive
import qualified Data.Set as S


label' :: (Graph gr) => gr a b -> Edge -> Maybe b
label' gr (u,v) = lookup v (lsuc gr u)

shortest_path :: (Real b, Graph gr) => gr a b -> Node -> Node -> Path
shortest_path graph node_1 node_2= sp node_1 node_2 graph


mkGraph' :: [LEdge b] -> Gr () b
mkGraph' es = mkGraph ns es
    where
        ns :: [LNode ()]
        ns = zip [1.. (fromIntegral . length) ns'] (repeat ())
            where ns' = S.toList . S.fromList $ concat (Prelude.map edge2nodes es)


edge2nodes :: LEdge b -> [Node]
edge2nodes (a,b,_) = [a,b]


eU :: [LEdge Double]
eU = [
     (2,1,1)
    ,(1,2,1)
    
    ,(1,4,1)
    ,(4,1,1)
    
    ,(2,3,1)
    ,(3,2,1)
    
    ,(3,4,1)
    ,(4,3,1)
    
    ,(4,5,1)
    ,(5,4,1)
    ]

eD :: [LEdge Double]
eD = [
     (2,1,1)
    
    ,(1,4,1)
    
    ,(2,3,1)
    
    ,(3,4,1)
    
    ,(4,5,1)
    ]

gU :: Gr () Double
gU = mkGraph' eU

-- > prettyPrint gU
-- 1:()->[(1,2),(1,4)]
-- 2:()->[(1,1),(1,3)]
-- 3:()->[(1,2),(1,4)]
-- 4:()->[(1,1),(1,3),(1,5)]
-- 5:()->[(1,4)]

-- Visual representation:
-- 
--    2
--   / \
--  1   3
--   \ /
--    4
--    |
--    5
-- 
-- 

gD :: Gr () Double
gD = mkGraph' eD

eD' :: [LEdge Double]
eD' = [
     (2,1,1)
    
    ,(1,4,1)
    
    ,(2,3,1)
    
    ,(3,4,1)
    
    ,(4,5,1)
    ,(5,6,1)
    ,(5,7,1)
    ,(6,7,1)
    ]
gD' :: Gr () Double
gD' = mkGraph' eD'


type Modularity = Double

rotate :: Int -> [a] -> [a]
rotate _ [] = []
rotate n xs = zipWith const (drop n (cycle xs)) xs

takeDrop :: Int -> [a] -> [[a]]
takeDrop n xs = [ (take n xs), drop n xs]

-- http://stackoverflow.com/questions/35388734/list-partitioning-implemented-recursively
separate :: [a] -> [[[a]]]
separate [] = [[]]
separate (x:xs) = let recur = separate xs
                      split = do
                        partition <- recur
                        return $ [x] : partition
                      noSplit = do
                        (y:ys) <- recur
                        return $ (x:y):ys
                  in split ++ noSplit

separate' xs = [ takeDrop t (rotate r xs) 
               | t <- [1.. fromIntegral (length xs) - 1 ]
               , r <- [0.. fromIntegral (length xs)     ]
               ]


gpartition :: DynGraph gr => gr a b -> [[[Node]]]
gpartition gr = separate' (nodes gr)

modularities :: DynGraph gr => gr a b -> [[Node]] -> Double
modularities gr xs = sum $ Prelude.map (\y -> modularity gr y) xs


compareModularity :: DynGraph gr => gr a b -> [[Node]] -> [[Node]] -> Ordering
compareModularity gr xs ys
    | modularities gr xs < modularities gr ys = LT
    | modularities gr xs > modularities gr ys = GT
    | otherwise = EQ


bestPartition :: DynGraph gr => gr a b -> [[Node]]
bestPartition gr = maximumBy (compareModularity gr) $ gpartition gr


modularity :: DynGraph gr => gr a b -> [Node] -> Double
modularity gr ns = coverage - edgeDensity
    where
        coverage :: Double
        coverage  = sizeSubGraph / sizeAllGraph 
            where
                sizeSubGraph :: Double
                sizeSubGraph = fromIntegral ( size $ subgraph ns gr )

                sizeAllGraph :: Double
                sizeAllGraph = fromIntegral (size gr)

        edgeDensity :: Double
        edgeDensity = (sum (Prelude.map (\node -> (degree node) / links ) ns)) ** 2
            where
                degree :: Node -> Double
                degree node = fromIntegral (deg gr node)

                links :: Double
                links       = fromIntegral (2 * (size gr))

